﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 메인 폼
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 육각형 너비
        /// </summary>
        private const float HEXAGON_WIDTH = 50;

        /// <summary>
        /// 육각형 위치 리스트
        /// </summary>
        private List<PointF> hexagonPointList = new List<PointF>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainForm()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            this.pictureBox.Resize     += pictureBox_Resize;
            this.pictureBox.Paint      += pictureBox_Paint;
            this.pictureBox.MouseMove  += pictureBox_MouseMove;
            this.pictureBox.MouseClick += pictureBox_MouseClick;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 픽처 박스 크기 조정시 처리하기 - pictureBox_Resize(sender, e)

        /// <summary>
        /// 픽처 박스 크기 조정시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_Resize(object sender, EventArgs e)
        {
            this.pictureBox.Refresh();
        }

        #endregion
        #region 픽처 박스 페인트시 처리하기 - pictureBox_Paint(sender, e)

        /// <summary>
        /// 픽처 박스 페인트시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;

            foreach(PointF hexagonPoint in this.hexagonPointList)
            {
                e.Graphics.FillPolygon
                (
                    Brushes.LightBlue,
                    GetHexagonPointArray
                    (
                        HEXAGON_WIDTH,
                        hexagonPoint.X,
                        hexagonPoint.Y
                    )
                );
            }

            DrawHexagonGrid
            (
                e.Graphics,
                Pens.Black,
                this.pictureBox.ClientSize.Width,
                this.pictureBox.ClientSize.Height,
                HEXAGON_WIDTH
            );
        }

        #endregion
        #region 픽처 박스 마우스 이동시 처리하기 - pictureBox_MouseMove(sender, e)

        /// <summary>
        /// 픽처 박스 마우스 이동시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            PointF hexagonPoint = GetHexagonPoint(e.X, e.Y, HEXAGON_WIDTH);

            Text = $"({hexagonPoint.Y}, {hexagonPoint.X})";
        }

        #endregion
        #region 픽처 박스 마우스 클릭시 처리하기 - pictureBox_MouseClick(sender, e)

        /// <summary>
        /// 픽처 박스 마우스 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_MouseClick(object sender, MouseEventArgs e)
        {
            PointF hexagonPoint = GetHexagonPoint(e.X, e.Y, HEXAGON_WIDTH);

            this.hexagonPointList.Add(hexagonPoint);

            this.pictureBox.Refresh();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 육각형 높이 구하기 - GetHexgonHeight(width)

        /// <summary>
        /// 육각형 높이 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <returns>육각형 높이</returns>
        private float GetHexgonHeight(float width)
        {
            return (float)(4 * (width / 2 / Math.Sqrt(3)));
        }

        #endregion
        #region 육각형 포인트 배열 구하기 - GetHexagonPointArray(width, row, column)

        /// <summary>
        /// 육각형 포인트 배열 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="row">행</param>
        /// <param name="column">열</param>
        /// <returns>육각형 포인트 배열</returns>
        private PointF[] GetHexagonPointArray(float width, float row, float column)
        {
            float height = GetHexgonHeight(width);
            float x      = width / 2;
            float y      = 0;

            x += column * width;

            if(row % 2 != 0)
            {
                x += width / 2;
            }

            y += row * (height * 0.75f);

            return new PointF[]
            {
                new PointF(x            , y                 ),
                new PointF(x + width / 2, y + height * 0.25f),
                new PointF(x + width / 2, y + height * 0.75f),
                new PointF(x            , y + height        ),
                new PointF(x - width / 2, y + height * 0.75f),
                new PointF(x - width / 2, y + height * 0.25f),
            };
        }

        #endregion
        #region 육각형 그리드 그리기 - DrawHexagonGrid(graphics, pen, maximumX, maximumY, height)

        /// <summary>
        /// 육각형 그리드 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="pen">펜</param>
        /// <param name="maximumX">최대 X</param>
        /// <param name="maximumY">최대 Y</param>
        /// <param name="height">높이</param>
        private void DrawHexagonGrid(Graphics graphics, Pen pen, float maximumX, float maximumY, float height)
        {
            for(int row = 0; ; row++)
            {
                PointF[] hexagonPointArray = GetHexagonPointArray(height, row, 0);

                if(hexagonPointArray[3].Y > maximumY)
                {
                    break;
                }

                for(int column = 0; ; column++)
                {
                    hexagonPointArray = GetHexagonPointArray(height, row, column);

                    if(hexagonPointArray[1].X > maximumX)
                    {
                        break;
                    }

                    if(hexagonPointArray[3].Y <= maximumY)
                    {
                        graphics.DrawPolygon(pen, hexagonPointArray);
                    }
                }
            }
        }

        #endregion
        #region 육각형 위치 구하기 - GetHexagonPoint(x, y, width)

        /// <summary>
        /// 육각형 위치 구하기
        /// </summary>
        /// <param name="x">X</param>
        /// <param name="y">Y</param>
        /// <param name="width">너비</param>
        /// <returns>육각형 위치</returns>
        private PointF GetHexagonPoint(float x, float y, float width)
        {
            float row;
            float column;

            float height = GetHexgonHeight(width);

            row = (int)(y / (height * 0.75f));

            if(row % 2 == 0)
            {
                column = (int)Math.Floor(x / width);
            }
            else
            {
                column = (int)Math.Floor((x - width / 2) / width);
            }

            float testX = column * width;
            float testY = row * height * 0.75f;

            if(row % 2 != 0)
            {
                testX += width / 2;
            }

            bool isLeft  = false;
            bool isRight = false;

            float deltaY = y - testY;

            if(deltaY < height / 4)
            {
                float deltaX = x - (testX + width / 2);

                if(deltaY < 0.001)
                {
                    if(deltaY < 0)
                    {
                        isLeft = true;
                    }

                    if(deltaY > 0)
                    {
                        isRight = true;
                    }
                }
                else if(deltaX < 0)
                {
                    if(-deltaX / deltaY > Math.Sqrt(3))
                    {
                        isLeft = true;
                    }
                }
                else
                {
                    if(deltaX / deltaY > Math.Sqrt(3))
                    {
                        isRight = true;
                    }
                }
            }

            if(isLeft)
            {
                if(row % 2 == 0)
                {
                    column--;
                }

                row--;
            }
            else if(isRight)
            {
                if(row % 2 != 0)
                {
                    column++;
                }

                row--;
            }

            return new PointF(row, column);
        }

        #endregion
    }
}